#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "pop3_session.h"
#include "authenticate.h"
#include "lock_maildrop.h"
#include "rw.h"
#include "process.h"
#include "signal.h"
#include "ssl.h"

#include <grp.h>
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>

#if HAVE_LIBSSL
# include <openssl/rsa.h>
# include <openssl/crypto.h>
# include <openssl/pem.h>
# include <openssl/ssl.h>
# include <openssl/err.h>
#endif

#define STATE_AUTH	1
#define STATE_TRANS	2
#define STATE_UPDATE	3

#define MAXLINE 4096

#define CERTF "/etc/akpop3d/cert.pem"
#define KEYF "/etc/akpop3d/key.pem"

static char * mdl;

static void do_remove_lock(void) {
  remove_lock(mdl);
}

static void sig_handler(int signo) {
  remove_lock(mdl);
  syslog(LOG_INFO,"Caught signal %d.",signo);
  exit(EXIT_FAILURE);
}

void pop3_session(int fd) {
  int state = STATE_AUTH;
  char user[MAXLINE+1], pass[MAXLINE+1], line[MAXLINE+1];
  char * userx, * passx;
  char * maildrop;
  int modified = 0;
  struct passwd * u_inf;
  struct group * g_inf;
#if HAVE_LIBSSL
  int rc;
  SSL_CTX * ctx = NULL;
  SSL * ssl = NULL;
  SSL_METHOD * meth = NULL;

  if (use_ssl()) {
    SSL_load_error_strings();
    SSLeay_add_ssl_algorithms();
    meth = SSLv23_server_method();
    ctx = SSL_CTX_new(meth);
    if (ctx==NULL) {
      syslog(LOG_ERR,"Problems with SSL.");
      exit(EXIT_FAILURE);
    }

    if (SSL_CTX_use_certificate_file(ctx,CERTF,SSL_FILETYPE_PEM) <= 0) {
      syslog(LOG_ERR,"Problems with CERTF.");
      exit(EXIT_FAILURE);
    }

    if (SSL_CTX_use_PrivateKey_file(ctx,KEYF,SSL_FILETYPE_PEM) <= 0) {
      syslog(LOG_ERR,"Problems with KEYF.");
      exit(EXIT_FAILURE);
    }

    ssl = SSL_new(ctx);
    if (ssl==NULL) {
      syslog(LOG_ERR,"Problems with SSL (#2).");
      exit(EXIT_FAILURE);
    }
    set_ssl_handle(ssl);

    SSL_set_fd(ssl,fd);
    rc = SSL_accept(ssl);
    if (rc==-1) {
      syslog(LOG_ERR,"SSL handshake problems.");
      exit(EXIT_FAILURE);
    }
  }
#endif

  /* I'm ready */
  write_line(fd, "+OK I'm ready.\r\n");
begin_auth:

  /* first read username */
  read_line(fd,user,MAXLINE);
  if (strncasecmp(user,"QUIT",4)==0) {
    write_line(fd,"+OK Bye.\r\n");
    exit(EXIT_SUCCESS);
  }

  /* if input was not username, repeat this step */
  if (strncasecmp(user,"USER ",5)!=0) {
    write_line(fd,"-ERR need USER command first.\r\n");
    goto begin_auth;
  }

  /* username is OK */
  write_line(fd,"+OK user name accepted.\r\n");

  /* now read password */
  read_line(fd,pass,MAXLINE);
  if (strncasecmp(pass,"QUIT",4)==0) {
    write_line(fd,"+OK Bye.\r\n");
    exit(EXIT_SUCCESS);
  }

  /* if input was not password, let user enter another
   * username and a password.
   */
  if (strncasecmp(pass,"PASS ",5)!=0) {
    write_line(fd,"-ERR need PASS command.\r\n");
    read_line(fd,line,MAXLINE);
    while (strncasecmp(line,"PASS ", 5)!=0) {
      if (strncasecmp(line,"USER ",5)==0) {
        strncpy(user,line,MAXLINE);
        write_line(fd,"+OK now PASS, please.\r\n");
      } else if (strncasecmp(line,"QUIT ",5)==0) {
        write_line(fd,"+OK Bye.\r\n");
        exit(EXIT_SUCCESS);
      } else {
        write_line(fd,"-ERR I currently want USER or PASS.\r\n");
      }
      read_line(fd,line,MAXLINE);
    }
    strncpy(pass,line,MAXLINE);
  }

  /*
   * we both have username and password, we can now
   * authenticate.
   */
  if (authenticate(user+5,pass+5)==0) {
    write_line(fd,"-ERR authentication failed.\r\n");
    syslog(LOG_WARNING,"Failed login. Aborting session.");
    exit(EXIT_FAILURE);
  }

#define clean_up(str) if (index(str,'\r') != NULL) { \
                        str[index(str,'\r')-str] = '\0'; \
                      } else if (index(str,'\n') != NULL) { \
                        str[index(str,'\n')-str] = '\0'; \
                      }

  clean_up(user);
  userx = user + 5;
  clean_up(pass);
  passx = pass + 5;

  /* we've reached transaction state */
  state = STATE_TRANS;

  maildrop = malloc(strlen("/var/mail/")+1+strlen(userx));
  if (maildrop==NULL) {
    write_line(fd,"-ERR can't lock maildrop .\r\n");
    syslog(LOG_ERR,"virtual memory exhausted: %m");
    exit(EXIT_FAILURE);
  }
  strcpy(maildrop,"/var/mail/");
  strcat(maildrop,userx);

  mdl = strdup(maildrop);

  u_inf = getpwnam(userx);
  if (u_inf==NULL) {
    syslog(LOG_ERR,"Impossible! getpwnam() fails for a user previously authorized: %m");
    write_line(fd,"-ERR user doesn't exist.\r\n");
    exit(EXIT_FAILURE);
  }

  g_inf = getgrnam("mail");
  if (g_inf==NULL) {
    syslog(LOG_ERR,"You need a group mail, but you don't have one.");
    write_line(fd,"-ERR group mail doesn't exist.\r\n");
    exit(EXIT_FAILURE);
  }
  if (setegid(g_inf->gr_gid)!=0) {
    syslog(LOG_ERR,"Abortin, since I can't setegid(%d): %m",g_inf->gr_gid);
    write_line(fd,"-ERR unexpected problems with group mails.\r\n");
    exit(EXIT_FAILURE);
  }

  if (lock_maildrop(maildrop,u_inf->pw_uid,g_inf->gr_gid)==0) {
    write_line(fd,"-ERR can't lock maildrop.\r\n");
    syslog(LOG_ERR,"Unable to lock maildrop %s: %m",maildrop);
    exit(EXIT_FAILURE);
  }

  if (seteuid(u_inf->pw_uid)!=0) {
    syslog(LOG_ERR,"Aborting, since I can't seteuid(%d): %m",u_inf->pw_uid);
    write_line(fd,"-ERR unexpected user authentication problems.\r\n");
    exit(EXIT_FAILURE);
  }

  if (atexit(do_remove_lock)!=0) {
    syslog(LOG_WARNING,"do_remove_lock couldn't be registered (%m) There'll probably be problems with lock files.");
  }

  Signal(SIGPIPE,sig_handler);

  if (process_mails(maildrop)==0) {
    write_line(fd,"-ERR internal mailbox problem.\r\n");
    do_remove_lock();
    exit(EXIT_FAILURE);
  }

  write_line(fd,"+OK You're authenticated and maildrop is locked. Let's go!\r\n");

  read_line(fd,line,MAXLINE);
  while (strncasecmp(line,"QUIT",4)!=0) {
    if (strncasecmp(line,"STAT",4)==0) {
      long msg, size;
      get_stats(&msg,&size);
      snprintf(line,MAXLINE,"+OK %ld %ld\r\n",msg,size);
      write_line(fd,line);
    } else if (strncasecmp(line,"LIST",4)==0) {
      do_list(fd,line); /* parses itself */
    } else if (strncasecmp(line,"RETR",4)==0) {
      retrieve_msg(fd,line); /* parses itself */
    } else if (strncasecmp(line,"DELE",4)==0) {
      delete_msg(fd,line); /* parses itself */
      modified = 1;
    } else if (strncasecmp(line,"NOOP",4)==0) {
      write_line(fd,"+OK\r\n");
      /* do nothing */
    } else if (strncasecmp(line,"RSET",4)==0) {
      reset_msg();
      modified = 0;
      write_line(fd,"+OK\r\n");
    } else if (strncasecmp(line,"TOP ",4)==0) {
      show_top_msg(fd,line);
    } else {
      write_line(fd,"-ERR Sorry, don't understand.\r\n");
    }
    read_line(fd,line,MAXLINE);
  }

  write_line(fd,"+OK Bye.\r\n");
  close(fd);
#if HAVE_LIBSSL
  if (use_ssl()) {
    SSL_free(ssl);
    SSL_CTX_free(ctx);
  }
#endif

  if (modified) {
    do_update(maildrop);
  }
  remove_lock(maildrop);
  free(maildrop);
}
